#!/usr/local/bin/python3.13

from __future__ import print_function

import sys
import optparse
import socket
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse # Python 2.7 required for Literal IPv6 Addresses
import astdicts
import astconfigparser

PREFIX = 'pjsip_'
QUIET = False

###############################################################################
### some utility functions
###############################################################################


def section_by_type(section, pjsip, type):
    """Finds a section based upon the given type, adding it if not found."""
    def __find_dict(mdicts, key, val):
        """Given a list of multi-dicts, return the multi-dict that contains
           the given key/value pair."""

        def found(d):
            return key in d and val in d[key]

        try:
            return [d for d in mdicts if found(d)][0]
        except IndexError:
            raise LookupError("Dictionary not located for key = %s, value = %s"
                              % (key, val))

    try:
        return __find_dict(pjsip.section(section), 'type', type)
    except LookupError:
        # section for type doesn't exist, so add
        sect = pjsip.add_section(section)
        sect['type'] = type
        return sect


def ignore(key=None, val=None, section=None, pjsip=None,
           nmapped=None, type='endpoint'):
    """Ignore a key and mark it as mapped"""


def set_value(key=None, val=None, section=None, pjsip=None,
              nmapped=None, type='endpoint'):
    """Sets the key to the value within the section in pjsip.conf"""
    def _set_value(k, v, s, r, n):
        set_value(key if key else k, v, s, r, n, type)

    # if no value or section return the set_value
    # function with the enclosed key and type
    if not val and not section:
        return _set_value

    # otherwise try to set the value
    section_by_type(section, pjsip, type)[key] = \
        val[0] if isinstance(val, list) else val


def merge_value(key=None, val=None, section=None, pjsip=None,
                nmapped=None, type='endpoint', section_to=None,
                key_to=None):
    """Merge values from the given section with those from the default."""
    def _merge_value(k, v, s, r, n):
        merge_value(key if key else k, v, s, r, n, type, section_to, key_to)

    # if no value or section return the merge_value
    # function with the enclosed key and type
    if not val and not section:
        return _merge_value

    # should return a single value section list
    try:
        sect = sip.section(section)[0]
    except LookupError:
        sect = sip.default(section)[0]
    # for each merged value add it to pjsip.conf
    for i in sect.get_merged(key):
        set_value(key_to if key_to else key, i,
                  section_to if section_to else section,
                  pjsip, nmapped, type)

def merge_codec_value(key=None, val=None, section=None, pjsip=None,
                nmapped=None, type='endpoint', section_to=None,
                key_to=None):
    """Merge values from allow/deny with those from the default. Special treatment for all"""
    def _merge_codec_value(k, v, s, r, n):
        merge_codec_value(key if key else k, v, s, r, n, type, section_to, key_to)

    # if no value or section return the merge_codec_value
    # function with the enclosed key and type
    if not val and not section:
        return _merge_codec_value

    if key == 'allow':
        try:
            disallow = sip.get(section, 'disallow')[0]
            if disallow == 'all':
                #don't inherit
                for i in sip.get(section, 'allow'):
                    set_value(key, i, section, pjsip, nmapped, type)
            else:
                merge_value(key, val, section, pjsip, nmapped, type, section_to, key_to)
        except LookupError:
            print("lookup error", file=sys.stderr)
            merge_value(key, val, section, pjsip, nmapped, type, section_to, key_to)
            return
    elif key == 'disallow':
        try:
            allow = sip.get(section, 'allow')[0]
            if allow == 'all':
                #don't inherit
                for i in sip.get(section, 'disallow'):
                    set_value(key, i, section, pjsip, nmapped, type)
            else:
                merge_value(key, val, section, pjsip, nmapped, type, section_to, key_to)
        except LookupError:
            merge_value(key, val, section, pjsip, nmapped, type, section_to, key_to)
            return
    else:
        merge_value(key, val, section, pjsip, nmapped, type, section_to, key_to)


def non_mapped(nmapped):
    """Write non-mapped sip.conf values to the non-mapped object"""
    def _non_mapped(section, key, val):
        """Writes a non-mapped value from sip.conf to the non-mapped object."""
        if section not in nmapped:
            nmapped[section] = astconfigparser.Section()
            if isinstance(val, list):
                for v in val:
                    # since coming from sip.conf we can assume
                    # single section lists
                    nmapped[section][0][key] = v
            else:
                nmapped[section][0][key] = val
    return _non_mapped

###############################################################################
### mapping functions -
###      define f(key, val, section) where key/val are the key/value pair to
###      write to given section in pjsip.conf
###############################################################################


def set_dtmfmode(key, val, section, pjsip, nmapped):
    """
    Sets the dtmfmode value.  If value matches allowable option in pjsip
    then map it, otherwise set it to none.
    """
    key = 'dtmf_mode'
    # available pjsip.conf values: rfc4733, inband, info, none
    if val == 'inband' or val == 'info' or val == 'auto':
        set_value(key, val, section, pjsip, nmapped)
    elif val == 'rfc2833':
        set_value(key, 'rfc4733', section, pjsip, nmapped)
    else:
        nmapped(section, key, val + " ; did not fully map - set to none")
        set_value(key, 'none', section, pjsip, nmapped)


def setup_udptl(section, pjsip, nmapped):
    """Sets values from udptl into the appropriate pjsip.conf options."""
    try:
        val = sip.get(section, 't38pt_udptl')[0]
    except LookupError:
        try:
            val = sip.get('general', 't38pt_udptl')[0]
        except LookupError:
            return

    ec = 'none'
    if 'yes' in val:
        set_value('t38_udptl', 'yes', section, pjsip, nmapped)
    if 'no' in val:
        set_value('t38_udptl', 'no', section, pjsip, nmapped)
    if 'redundancy' in val:
        ec = 'redundancy'
    if 'fec' in val:
        ec = 'fec'
    set_value('t38_udptl_ec', ec, section, pjsip, nmapped)

def from_nat(key, val, section, pjsip, nmapped):
    """Sets values from nat into the appropriate pjsip.conf options."""
    # nat from sip.conf can be comma separated list of values:
    # yes/no, [auto_]force_rport, [auto_]comedia
    if 'yes' in val:
        set_value('rtp_symmetric', 'yes', section, pjsip, nmapped)
        set_value('rewrite_contact', 'yes', section, pjsip, nmapped)
    if 'comedia' in val:
        set_value('rtp_symmetric', 'yes', section, pjsip, nmapped)
    if 'force_rport' in val:
        set_value('force_rport', 'yes', section, pjsip, nmapped)
        set_value('rewrite_contact', 'yes', section, pjsip, nmapped)


def set_timers(key, val, section, pjsip, nmapped):
    """
    Sets the timers in pjsip.conf from the session-timers option
    found in sip.conf.
    """
    # pjsip.conf values can be yes/no, required, always
    # 'required' is a new feature of chan_pjsip, which rejects
    #            all SIP clients not supporting Session Timers
    # 'Accept' is the default value of chan_sip and maps to 'yes'
    # chan_sip ignores the case, for example 'session-timers=Refuse'
    val = val.lower()
    if val == 'originate':
        set_value('timers', 'always', section, pjsip, nmapped)
    elif val == 'refuse':
        set_value('timers', 'no', section, pjsip, nmapped)
    else:
        set_value('timers', 'yes', section, pjsip, nmapped)


def set_direct_media(key, val, section, pjsip, nmapped):
    """
    Maps values from the sip.conf comma separated direct_media option
    into pjsip.conf direct_media options.
    """
    if 'yes' in val:
        set_value('direct_media', 'yes', section, pjsip, nmapped)
    if 'update' in val:
        set_value('direct_media_method', 'update', section, pjsip, nmapped)
    if 'outgoing' in val:
        set_value('directed_media_glare_mitigation', 'outgoing', section,
                  pjsip, nmapped)
    if 'nonat' in val:
        set_value('disable_directed_media_on_nat', 'yes', section, pjsip,
                  nmapped)
    if 'no' in val:
        set_value('direct_media', 'no', section, pjsip, nmapped)


def from_sendrpid(key, val, section, pjsip, nmapped):
    """Sets the send_rpid/pai values in pjsip.conf."""
    if val == 'yes' or val == 'rpid':
        set_value('send_rpid', 'yes', section, pjsip, nmapped)
    elif val == 'pai':
        set_value('send_pai', 'yes', section, pjsip, nmapped)


def set_media_encryption(key, val, section, pjsip, nmapped):
    """Sets the media_encryption value in pjsip.conf"""
    try:
        dtls = sip.get(section, 'dtlsenable')[0]
        if dtls == 'yes':
            # If DTLS is enabled, then that overrides SDES encryption.
            return
    except LookupError:
        pass

    if val == 'yes':
        set_value('media_encryption', 'sdes', section, pjsip, nmapped)


def from_recordfeature(key, val, section, pjsip, nmapped):
    """
    If record on/off feature is set to automixmon then set
    one_touch_recording, otherwise it can't be mapped.
    """
    set_value('one_touch_recording', 'yes', section, pjsip, nmapped)
    set_value(key, val, section, pjsip, nmapped)

def set_record_on_feature(key, val, section, pjsip, nmapped):
    """Sets the record_on_feature in pjsip.conf"""
    from_recordfeature('record_on_feature', val, section, pjsip, nmapped)

def set_record_off_feature(key, val, section, pjsip, nmapped):
    """Sets the record_off_feature in pjsip.conf"""
    from_recordfeature('record_off_feature', val, section, pjsip, nmapped)

def from_progressinband(key, val, section, pjsip, nmapped):
    """Sets the inband_progress value in pjsip.conf"""
    # progressinband can = yes/no/never
    if val == 'never':
        val = 'no'
    set_value('inband_progress', val, section, pjsip, nmapped)


def build_host(config, host, section='general', port_key=None):
    """
    Returns a string composed of a host:port. This assumes that the host
    may have a port as part of the initial value. The port_key overrides
    a port in host, see parameter 'bindport' in chan_sip.
    """
    try:
        socket.inet_pton(socket.AF_INET6, host)
        if not host.startswith('['):
            # SIP URI will need brackets.
            host = '[' + host + ']'
    except socket.error:
        pass

    # Literal IPv6 (like [::]), IPv4, or hostname
    # does not work for IPv6 without brackets; case catched above
    url = urlparse('sip://' + host)

    if port_key:
        try:
            port = config.get(section, port_key)[0]
            host = url.hostname # no port, but perhaps no brackets
            try:
                socket.inet_pton(socket.AF_INET6, host)
                if not host.startswith('['):
                    # SIP URI will need brackets.
                    host = '[' + host + ']'
            except socket.error:
                pass
            return host + ':' + port
        except LookupError:
            pass

    # Returns host:port, in brackets if required
    # TODO Does not compress IPv6, for example 0:0:0:0:0:0:0:0 should get [::]
    return url.netloc


def from_host(key, val, section, pjsip, nmapped):
    """
    Sets contact info in an AOR section in pjsip.conf using 'host'
    and 'port' data from sip.conf
    """
    # all aors have the same name as the endpoint so makes
    # it easy to set endpoint's 'aors' value
    set_value('aors', section, section, pjsip, nmapped)
    if val == 'dynamic':
        # Easy case. Just set the max_contacts on the aor and we're done
        set_value('max_contacts', 1, section, pjsip, nmapped, 'aor')
        return

    result = 'sip:'

    # More difficult case. The host will be either a hostname or
    # IP address and may or may not have a port specified. pjsip.conf
    # expects the contact to be a SIP URI.

    user = None

    try:
        user = sip.multi_get(section, ['defaultuser', 'username'])[0]
        result += user + '@'
    except LookupError:
        # It's fine if there's no user name
        pass

    result += build_host(sip, val, section, 'port')

    set_value('contact', result, section, pjsip, nmapped, 'aor')


def from_mailbox(key, val, section, pjsip, nmapped):
    """
    Determines whether a mailbox configured in sip.conf should map to
    an endpoint or aor in pjsip.conf. If subscribemwi is true, then the
    mailboxes are set on an aor. Otherwise the mailboxes are set on the
    endpoint.
    """

    try:
        subscribemwi = sip.get(section, 'subscribemwi')[0]
    except LookupError:
        # No subscribemwi option means default it to 'no'
        subscribemwi = 'no'

    set_value('mailboxes', val, section, pjsip, nmapped, 'aor'
              if subscribemwi == 'yes' else 'endpoint')


def setup_auth(key, val, section, pjsip, nmapped):
    """
    Sets up authentication information for a specific endpoint based on the
    'secret' setting on a peer in sip.conf
    """
    set_value('username', section, section, pjsip, nmapped, 'auth')
    # In chan_sip, if a secret and an md5secret are both specified on a peer,
    # then in practice, only the md5secret is used. If both are encountered
    # then we build an auth section that has both an md5_cred and password.
    # However, the auth_type will indicate to authenticators to use the
    # md5_cred, so like with sip.conf, the password will be there but have
    # no purpose.
    if key == 'secret':
        set_value('password', val, section, pjsip, nmapped, 'auth')
    else:
        set_value('md5_cred', val, section, pjsip, nmapped, 'auth')
        set_value('auth_type', 'md5', section, pjsip, nmapped, 'auth')

    realms = [section]
    try:
        auths = sip.get('authentication', 'auth')
        for i in auths:
            user, at, realm = i.partition('@')
            realms.append(realm)
    except LookupError:
        pass

    realm_str = ','.join(realms)

    set_value('auth', section, section, pjsip, nmapped)
    set_value('outbound_auth', realm_str, section, pjsip, nmapped)


def setup_ident(key, val, section, pjsip, nmapped):
    """
    Examines the 'type' field for a sip.conf peer and creates an identify
    section if the type is either 'peer' or 'friend'. The identify section uses
    either the host or defaultip field of the sip.conf peer.
    """
    if val != 'peer' and val != 'friend':
        return

    try:
        ip = sip.get(section, 'host')[0]
    except LookupError:
        return

    if ip == 'dynamic':
        try:
            ip = sip.get(section, 'defaultip')[0]
        except LookupError:
            return

    set_value('endpoint', section, section, pjsip, nmapped, 'identify')
    set_value('match', ip, section, pjsip, nmapped, 'identify')


def from_encryption_taglen(key, val, section, pjsip, nmapped):
    """Sets the srtp_tag32 option based on sip.conf encryption_taglen"""
    if val == '32':
        set_value('srtp_tag_32', 'yes', section, pjsip, nmapped)


def from_dtlsenable(key, val, section, pjsip, nmapped):
    """Optionally sets media_encryption=dtls based on sip.conf dtlsenable"""
    if val == 'yes':
        set_value('media_encryption', 'dtls', section, pjsip, nmapped)

###############################################################################

# options in pjsip.conf on an endpoint that have no sip.conf equivalent:
# type, 100rel, trust_id_outbound, aggregate_mwi, connected_line_method

# known sip.conf peer keys that can be mapped to a pjsip.conf section/key
peer_map = [
    # sip.conf option      mapping function     pjsip.conf option(s)
    ###########################################################################
    ['context',            set_value],
    ['dtmfmode',           set_dtmfmode],
    ['disallow',           merge_codec_value],
    ['allow',              merge_codec_value],
    ['nat',                from_nat],            # rtp_symmetric, force_rport,
                                                 # rewrite_contact
    ['rtptimeout',         set_value('rtp_timeout')],
    ['icesupport',         set_value('ice_support')],
    ['autoframing',        set_value('use_ptime')],
    ['outboundproxy',      set_value('outbound_proxy')],
    ['mohsuggest',         set_value('moh_suggest')],
    ['session-timers',     set_timers],          # timers
    ['session-minse',      set_value('timers_min_se')],
    ['session-expires',    set_value('timers_sess_expires')],
    # identify_by ?
    ['canreinvite',        set_direct_media],    # direct_media alias
    ['directmedia',        set_direct_media],    # direct_media
                                                 # direct_media_method
                                                 # directed_media_glare_mitigation
                                                 # disable_directed_media_on_nat
    ['callerid',           set_value],           # callerid
    ['callingpres',        set_value('callerid_privacy')],
    ['cid_tag',            set_value('callerid_tag')],
    ['trustrpid',          set_value('trust_id_inbound')],
    ['sendrpid',           from_sendrpid],       # send_pai, send_rpid
    ['send_diversion',     set_value],
    ['encryption',         set_media_encryption],
    ['avpf',               set_value('use_avpf')],
    ['recordonfeature',    set_record_on_feature],  # automixon
    ['recordofffeature',   set_record_off_feature],  # automixon
    ['progressinband',     from_progressinband], # in_band_progress
    ['callgroup',          set_value('call_group')],
    ['pickupgroup',        set_value('pickup_group')],
    ['namedcallgroup',     set_value('named_call_group')],
    ['namedpickupgroup',   set_value('named_pickup_group')],
    ['allowtransfer',      set_value('allow_transfer')],
    ['fromuser',           set_value('from_user')],
    ['fromdomain',         set_value('from_domain')],
    ['mwifrom',            set_value('mwi_from_user')],
    ['tos_audio',          set_value],
    ['tos_video',          set_value],
    ['cos_audio',          set_value],
    ['cos_video',          set_value],
    ['sdpowner',           set_value('sdp_owner')],
    ['sdpsession',         set_value('sdp_session')],
    ['tonezone',           set_value('tone_zone')],
    ['language',           set_value],
    ['allowsubscribe',     set_value('allow_subscribe')],
    ['subscribecontext',   set_value('subscribe_context')],
    ['subminexpiry',       set_value('sub_min_expiry')],
    ['rtp_engine',         set_value],
    ['mailbox',            from_mailbox],
    ['busylevel',          set_value('device_state_busy_at')],
    ['secret',             setup_auth],
    ['md5secret',          setup_auth],
    ['type',               setup_ident],
    ['dtlsenable',         from_dtlsenable],
    ['dtlsverify',         set_value('dtls_verify')],
    ['dtlsrekey',          set_value('dtls_rekey')],
    ['dtlscertfile',       set_value('dtls_cert_file')],
    ['dtlsprivatekey',     set_value('dtls_private_key')],
    ['dtlscipher',         set_value('dtls_cipher')],
    ['dtlscafile',         set_value('dtls_ca_file')],
    ['dtlscapath',         set_value('dtls_ca_path')],
    ['dtlssetup',          set_value('dtls_setup')],
    ['encryption_taglen',  from_encryption_taglen],
    ['setvar',             ignore],

############################ maps to an aor ###################################

    ['host',               from_host],           # contact, max_contacts
    ['qualifyfreq',        set_value('qualify_frequency', type='aor')],
    ['maxexpiry',          set_value('maximum_expiration', type='aor')],
    ['minexpiry',          set_value('minimum_expiration', type='aor')],
    ['defaultexpiry',      set_value('default_expiration', type='aor')],

############################# maps to auth#####################################
#        type = auth
#        username
#        password
#        md5_cred
#        realm
#        nonce_lifetime
#        auth_type
######################### maps to acl/security ################################

    ['permit',             merge_value(type='acl', section_to='acl')],
    ['deny',               merge_value(type='acl', section_to='acl')],
    ['acl',                merge_value(type='acl', section_to='acl')],
    ['contactpermit',      merge_value(type='acl', section_to='acl', key_to='contact_permit')],
    ['contactdeny',        merge_value(type='acl', section_to='acl', key_to='contact_deny')],
    ['contactacl',         merge_value(type='acl', section_to='acl', key_to='contact_acl')],

########################### maps to transport #################################
#        type = transport
#        protocol
#        bind
#        async_operations
#        ca_list_file
#        ca_list_path
#        cert_file
#        privkey_file
#        password
#        external_signaling_address - externip & externhost
#        external_signaling_port
#        external_media_address
#        domain
#        verify_server
#        verify_client
#        require_client_cert
#        method
#        cipher
#        localnet
######################### maps to domain_alias ################################
#        type = domain_alias
#        domain
######################### maps to registration ################################
#        type = registration
#        server_uri
#        client_uri
#        contact_user
#        transport
#        outbound_proxy
#        expiration
#        retry_interval
#        max_retries
#        auth_rejection_permanent
#        outbound_auth
########################### maps to identify ##################################
#        type = identify
#        endpoint
#        match
]


def split_hostport(addr):
    """
    Given an address in the form 'host:port' separate the host and port
    components.
    Returns a two-tuple of strings, (host, port). If no port is present in the
    string, then the port section of the tuple is None.
    """
    try:
        socket.inet_pton(socket.AF_INET6, addr)
        if not addr.startswith('['):
            return (addr, None)
    except socket.error:
        pass

    # Literal IPv6 (like [::]), IPv4, or hostname
    # does not work for IPv6 without brackets; case catched above
    url = urlparse('sip://' + addr)
    # TODO Does not compress IPv6, for example 0:0:0:0:0:0:0:0 should get [::]
    return (url.hostname, url.port)


def set_transport_common(section, sip, pjsip, protocol, nmapped):
    """
    sip.conf has several global settings that in pjsip.conf apply to individual
    transports. This function adds these global settings to each individual
    transport.

    The settings included are:
    externaddr (or externip)
    externhost
    externtcpport for TCP
    externtlsport for TLS
    localnet
    tos_sip
    cos_sip
    """
    try:
        extern_addr = sip.multi_get('general', ['externaddr', 'externip',
                                                'externhost'])[0]
        host, port = split_hostport(extern_addr)
        try:
            port = sip.get('general', 'extern' + protocol + 'port')[0]
        except LookupError:
            pass
        set_value('external_media_address', host, section, pjsip,
                  nmapped, 'transport')
        set_value('external_signaling_address', host, section, pjsip,
                  nmapped, 'transport')
        if port:
            set_value('external_signaling_port', port, section, pjsip,
                      nmapped, 'transport')
    except LookupError:
        pass

    try:
        merge_value('localnet', sip.get('general', 'localnet')[0], 'general',
                    pjsip, nmapped, 'transport', section, "local_net")
    except LookupError:
        # No localnet options configured. Move on.
        pass

    try:
        set_value('tos', sip.get('general', 'tos_sip')[0], section, pjsip,
                  nmapped, 'transport')
    except LookupError:
        pass

    try:
        set_value('cos', sip.get('general', 'cos_sip')[0], section, pjsip,
                  nmapped, 'transport')
    except LookupError:
        pass


def get_bind(sip, pjsip, protocol):
    """
    Given the protocol (udp, tcp, or tls), return
    - the bind address, like [::] or 0.0.0.0
    - name of the section to be created
    """
    section = 'transport-' + protocol

    # UDP cannot be disabled in chan_sip
    if protocol != 'udp':
        try:
            enabled = sip.get('general', protocol + 'enable')[0]
        except LookupError:
            # No value means disabled by default. Don't create this transport
            return (None, section)
        if enabled != 'yes':
            return (None, section)

    try:
        bind = pjsip.get(section, 'bind')[0]
        # The first run created an transport already but this
        # server was not configured for IPv4/IPv6 Dual Stack
        return (None, section)
    except LookupError:
        pass

    try:
        bind = pjsip.get(section + '6', 'bind')[0]
        # The first run created an IPv6 transport, because
        # the server was configured with :: as bindaddr.
        # Now, re-use its port and create the IPv4 transport
        host, port = split_hostport(bind)
        bind = '0.0.0.0'
        if port:
            bind += ':' + str(port)
    except LookupError:
        # This is the first run, no transport in pjsip exists.
        try:
            bind = sip.get('general', protocol + 'bindaddr')[0]
        except LookupError:
            if protocol == 'udp':
                try:
                    bind = sip.get('general', 'bindaddr')[0]
                except LookupError:
                    bind = '0.0.0.0'
            else:
                try:
                    bind = pjsip.get('transport-udp6', 'bind')[0]
                except LookupError:
                    bind = pjsip.get('transport-udp', 'bind')[0]
                # Only TCP reuses host:port of UDP, others reuse just host
                if protocol == 'tls':
                    bind, port = split_hostport(bind)
        host, port = split_hostport(bind)
        if host == '::':
            section += '6'

    if protocol == 'udp':
        host = build_host(sip, bind, 'general', 'bindport')
    else:
        host = build_host(sip, bind)

    return (host, section)


def create_udp(sip, pjsip, nmapped):
    """
    Creates a 'transport-udp' section in the pjsip.conf file based
    on the following settings from sip.conf:

    bindaddr (or udpbindaddr)
    bindport
    """
    protocol = 'udp'
    bind, section = get_bind(sip, pjsip, protocol)

    set_value('protocol', protocol, section, pjsip, nmapped, 'transport')
    set_value('bind', bind, section, pjsip, nmapped, 'transport')
    set_transport_common(section, sip, pjsip, protocol, nmapped)


def create_tcp(sip, pjsip, nmapped):
    """
    Creates a 'transport-tcp' section in the pjsip.conf file based
    on the following settings from sip.conf:

    tcpenable
    tcpbindaddr (or bindaddr)
    """
    protocol = 'tcp'
    bind, section = get_bind(sip, pjsip, protocol)
    if not bind:
        return

    set_value('protocol', protocol, section, pjsip, nmapped, 'transport')
    set_value('bind', bind, section, pjsip, nmapped, 'transport')
    set_transport_common(section, sip, pjsip, protocol, nmapped)


def set_tls_cert_file(val, pjsip, section, nmapped):
    """Sets cert_file based on sip.conf tlscertfile"""
    set_value('cert_file', val, section, pjsip, nmapped,
              'transport')


def set_tls_private_key(val, pjsip, section, nmapped):
    """Sets privkey_file based on sip.conf tlsprivatekey or sslprivatekey"""
    set_value('priv_key_file', val, section, pjsip, nmapped,
              'transport')


def set_tls_cipher(val, pjsip, section, nmapped):
    """Sets cipher based on sip.conf tlscipher or sslcipher"""
    if val == 'ALL':
        return
    print('chan_sip ciphers do not match 1:1 with PJSIP ciphers.' \
          ' You should manually review and adjust this.', file=sys.stderr)

    set_value('cipher', val, section, pjsip, nmapped, 'transport')


def set_tls_cafile(val, pjsip, section, nmapped):
    """Sets ca_list_file based on sip.conf tlscafile"""
    set_value('ca_list_file', val, section, pjsip, nmapped,
              'transport')


def set_tls_capath(val, pjsip, section, nmapped):
    """Sets ca_list_path based on sip.conf tlscapath"""
    set_value('ca_list_path', val, section, pjsip, nmapped,
              'transport')


def set_tls_verifyclient(val, pjsip, section, nmapped):
    """Sets verify_client based on sip.conf tlsverifyclient"""
    set_value('verify_client', val, section, pjsip, nmapped,
              'transport')


def set_tls_verifyserver(val, pjsip, section, nmapped):
    """Sets verify_server based on sip.conf tlsdontverifyserver"""

    if val == 'no':
        set_value('verify_server', 'yes', section, pjsip, nmapped,
                  'transport')
    else:
        set_value('verify_server', 'no', section, pjsip, nmapped,
                  'transport')


def create_tls(sip, pjsip, nmapped):
    """
    Creates a 'transport-tls' section in pjsip.conf based on the following
    settings from sip.conf:

    tlsenable (or sslenable)
    tlsbindaddr (or sslbindaddr or bindaddr)
    tlsprivatekey (or sslprivatekey)
    tlscipher (or sslcipher)
    tlscafile
    tlscapath (or tlscadir)
    tlscertfile (or sslcert or tlscert)
    tlsverifyclient
    tlsdontverifyserver
    tlsclientmethod (or sslclientmethod)
    """
    protocol = 'tls'
    bind, section = get_bind(sip, pjsip, protocol)
    if not bind:
        return

    set_value('protocol', protocol, section, pjsip, nmapped, 'transport')
    set_value('bind', bind, section, pjsip, nmapped, 'transport')
    set_transport_common(section, sip, pjsip, protocol, nmapped)

    tls_map = [
        (['tlscertfile', 'sslcert', 'tlscert'], set_tls_cert_file),
        (['tlsprivatekey', 'sslprivatekey'], set_tls_private_key),
        (['tlscipher', 'sslcipher'], set_tls_cipher),
        (['tlscafile'], set_tls_cafile),
        (['tlscapath', 'tlscadir'], set_tls_capath),
        (['tlsverifyclient'], set_tls_verifyclient),
        (['tlsdontverifyserver'], set_tls_verifyserver)
    ]

    for i in tls_map:
        try:
            i[1](sip.multi_get('general', i[0])[0], pjsip, section, nmapped)
        except LookupError:
            pass

    try:
        method = sip.multi_get('general', ['tlsclientmethod',
                                           'sslclientmethod'])[0]
        if section != 'transport-' + protocol + '6':  # print only once
            print('In chan_sip, you specified the TLS version. With chan_sip,' \
                  ' this was just for outbound client connections. In' \
                  ' chan_pjsip, this value is for client and server. Instead,' \
                  ' consider not to specify \'tlsclientmethod\' for chan_sip' \
                  ' and \'method = sslv23\' for chan_pjsip.', file=sys.stderr)
    except LookupError:
        """
        OpenSSL emerged during the 90s. SSLv2 and SSLv3 were the only
        existing methods at that time. The OpenSSL project continued. And as
        of today (OpenSSL 1.0.2) this does not start SSLv2 and SSLv3 anymore
        but TLSv1.0 and v1.2. Or stated differently: This method should
        have been called 'method = secure' or 'method = automatic' back in
        the 90s. The PJProject did not realize this and uses 'tlsv1' as
        default when unspecified, which disables TLSv1.2. chan_sip used
        'sslv23' as default when unspecified, which gives TLSv1.0 and v1.2.
        """
        method = 'sslv23'
    set_value('method', method, section, pjsip, nmapped, 'transport')


def map_transports(sip, pjsip, nmapped):
    """
    Finds options in sip.conf general section pertaining to
    transport configuration and creates appropriate transport
    configuration sections in pjsip.conf.

    sip.conf only allows a single UDP transport, TCP transport,
    and TLS transport for each IP version. As such, the mapping
    into PJSIP can be made consistent by defining six sections:

    transport-udp6
    transport-udp
    transport-tcp6
    transport-tcp
    transport-tls6
    transport-tls

    To accommodate the default behaviors in sip.conf, we'll need to
    create the UDP transports first, followed by the TCP and TLS transports.
    """

    # First create a UDP transport. Even if no bind parameters were provided
    # in sip.conf, chan_sip would always bind to UDP 0.0.0.0:5060
    create_udp(sip, pjsip, nmapped)
    create_udp(sip, pjsip, nmapped)

    # TCP settings may be dependent on UDP settings, so do it second.
    create_tcp(sip, pjsip, nmapped)
    create_tcp(sip, pjsip, nmapped)
    create_tls(sip, pjsip, nmapped)
    create_tls(sip, pjsip, nmapped)


def map_auth(sip, pjsip, nmapped):
    """
    Creates auth sections based on entries in the authentication section of
    sip.conf. pjsip.conf section names consist of "auth_" followed by the name
    of the realm.
    """
    try:
        auths = sip.get('authentication', 'auth')
    except LookupError:
        return

    for i in auths:
        creds, at, realm = i.partition('@')
        if not at and not realm:
            # Invalid. Move on
            continue
        user, colon, secret = creds.partition(':')
        if not secret:
            user, sharp, md5 = creds.partition('#')
            if not md5:
                #Invalid. move on
                continue
        section = "auth_" + realm

        set_value('realm', realm, section, pjsip, nmapped, 'auth')
        set_value('username', user, section, pjsip, nmapped, 'auth')
        if secret:
            set_value('password', secret, section, pjsip, nmapped, 'auth')
        else:
            set_value('md5_cred', md5, section, pjsip, nmapped, 'auth')
            set_value('auth_type', 'md5', section, pjsip, nmapped, 'auth')


class Registration:
    """
    Class for parsing and storing information in a register line in sip.conf.
    """
    def __init__(self, line, retry_interval, max_attempts, outbound_proxy):
        self.retry_interval = retry_interval
        self.max_attempts = max_attempts
        self.outbound_proxy = outbound_proxy
        self.parse(line)

    def parse(self, line):
        """
        Initial parsing routine for register lines in sip.conf.

        This splits the line into the part before the host, and the part
        after the '@' symbol. These two parts are then passed to their
        own parsing routines
        """

        # register =>
        # [peer?][transport://]user[@domain][:secret[:authuser]]@host[:port][/extension][~expiry]

        prehost, at, host_part = line.rpartition('@')
        if not prehost:
            raise

        self.parse_host_part(host_part)
        self.parse_user_part(prehost)

    def parse_host_part(self, host_part):
        """
        Parsing routine for the part after the final '@' in a register line.
        The strategy is to use partition calls to peel away the data starting
        from the right and working to the left.
        """
        pre_expiry, sep, expiry = host_part.partition('~')
        pre_extension, sep, self.extension = pre_expiry.partition('/')
        self.host, sep, self.port = pre_extension.partition(':')

        self.expiry = expiry if expiry else '120'

    def parse_user_part(self, user_part):
        """
        Parsing routine for the part before the final '@' in a register line.
        The only mandatory part of this line is the user portion. The strategy
        here is to start by using partition calls to remove everything to
        the right of the user, then finish by using rpartition calls to remove
        everything to the left of the user.
        """
        self.peer = ''
        self.protocol = 'udp'
        protocols = ['udp', 'tcp', 'tls']
        for protocol in protocols:
            position = user_part.find(protocol + '://')
            if -1 < position:
                post_transport = user_part[position + 6:]
                self.peer, sep, self.protocol = user_part[:position + 3].rpartition('?')
                user_part = post_transport
                break

        colons = user_part.count(':')
        if (colons == 3):
            # :domainport:secret:authuser
            pre_auth, sep, port_auth = user_part.partition(':')
            self.domainport, sep, auth = port_auth.partition(':')
            self.secret, sep, self.authuser = auth.partition(':')
        elif (colons == 2):
            # :secret:authuser
            pre_auth, sep, auth = user_part.partition(':')
            self.secret, sep, self.authuser = auth.partition(':')
        elif (colons == 1):
            # :secret
            pre_auth, sep, self.secret = user_part.partition(':')
        elif (colons == 0):
            # No port, secret, or authuser
            pre_auth = user_part
        else:
            # Invalid setting
            raise

        self.user, sep, self.domain = pre_auth.partition('@')

    def write(self, pjsip, nmapped):
        """
        Write parsed registration data into a section in pjsip.conf

        Most of the data in self will get written to a registration section.
        However, there will also need to be an auth section created if a
        secret or authuser is present.

        General mapping of values:
        A combination of self.host and self.port is server_uri
        A combination of self.user, self.domain, and self.domainport is
          client_uri
        self.expiry is expiration
        self.extension is contact_user
        self.protocol will map to one of the mapped transports
        self.secret and self.authuser will result in a new auth section, and
          outbound_auth will point to that section.
        XXX self.peer really doesn't map to anything :(
        """

        section = 'reg_' + self.host

        set_value('retry_interval', self.retry_interval, section, pjsip,
                  nmapped, 'registration')
        set_value('max_retries', self.max_attempts, section, pjsip, nmapped,
                  'registration')
        if self.extension:
            set_value('contact_user', self.extension, section, pjsip, nmapped,
                      'registration')

        set_value('expiration', self.expiry, section, pjsip, nmapped,
                  'registration')

        if self.protocol == 'udp':
            set_value('transport', 'transport-udp', section, pjsip, nmapped,
                      'registration')
        elif self.protocol == 'tcp':
            set_value('transport', 'transport-tcp', section, pjsip, nmapped,
                      'registration')
        elif self.protocol == 'tls':
            set_value('transport', 'transport-tls', section, pjsip, nmapped,
                      'registration')

        auth_section = 'auth_reg_' + self.host

        if hasattr(self, 'secret') and self.secret:
            set_value('password', self.secret, auth_section, pjsip, nmapped,
                      'auth')
            set_value('username', self.authuser if hasattr(self, 'authuser')
                      else self.user, auth_section, pjsip, nmapped, 'auth')
            set_value('outbound_auth', auth_section, section, pjsip, nmapped,
                      'registration')

        client_uri = "sip:%s@" % self.user
        if self.domain:
            client_uri += self.domain
        else:
            client_uri += self.host

        if hasattr(self, 'domainport') and self.domainport:
            client_uri += ":" + self.domainport
        elif self.port:
            client_uri += ":" + self.port

        set_value('client_uri', client_uri, section, pjsip, nmapped,
                  'registration')

        server_uri = "sip:%s" % self.host
        if self.port:
            server_uri += ":" + self.port

        set_value('server_uri', server_uri, section, pjsip, nmapped,
                  'registration')

        if self.outbound_proxy:
            set_value('outboundproxy', self.outbound_proxy, section, pjsip,
                      nmapped, 'registration')


def map_registrations(sip, pjsip, nmapped):
    """
    Gathers all necessary outbound registration data in sip.conf and creates
    corresponding registration sections in pjsip.conf
    """
    try:
        regs = sip.get('general', 'register')
    except LookupError:
        return

    try:
        retry_interval = sip.get('general', 'registertimeout')[0]
    except LookupError:
        retry_interval = '20'

    try:
        max_attempts = sip.get('general', 'registerattempts')[0]
    except LookupError:
        max_attempts = '10'

    try:
        outbound_proxy = sip.get('general', 'outboundproxy')[0]
    except LookupError:
        outbound_proxy = ''

    for i in regs:
        reg = Registration(i, retry_interval, max_attempts, outbound_proxy)
        reg.write(pjsip, nmapped)


def map_setvars(sip, section, pjsip, nmapped):
    """
    Map all setvar in peer section to the appropriate endpoint set_var
    """
    try:
        setvars = sip.section(section)[0].get('setvar')
    except LookupError:
        return

    for setvar in setvars:
        set_value('set_var', setvar, section, pjsip, nmapped)


def map_peer(sip, section, pjsip, nmapped):
    """
    Map the options from a peer section in sip.conf into the appropriate
    sections in pjsip.conf
    """
    for i in peer_map:
        try:
            # coming from sip.conf the values should mostly be a list with a
            # single value.  In the few cases that they are not a specialized
            # function (see merge_value) is used to retrieve the values.
            i[1](i[0], sip.get(section, i[0])[0], section, pjsip, nmapped)
        except LookupError:
            pass  # key not found in sip.conf

    setup_udptl(section, pjsip, nmapped)

def find_non_mapped(sections, nmapped):
    """
    Determine sip.conf options that were not properly mapped to pjsip.conf
    options.
    """
    for section, sect in sections.iteritems():
        try:
            # since we are pulling from sip.conf this should always
            # be a single value list
            sect = sect[0]
            # loop through the section and store any values that were not
            # mapped
            for key in sect.keys(True):
                for i in peer_map:
                    if i[0] == key:
                        break
                else:
                    nmapped(section, key, sect[key])
        except LookupError:
            pass


def map_system(sip, pjsip, nmapped):
    section = 'system' # Just a label; you as user can change that
    type = 'system' # Not a label, therefore not the same as section

    try:
        user_agent = sip.get('general', 'useragent')[0]
        set_value('user_agent', user_agent, 'global', pjsip, nmapped, 'global')
    except LookupError:
        pass


    try:
        sipdebug = sip.get('general', 'sipdebug')[0]
        set_value('debug', sipdebug, 'global', pjsip, nmapped, 'global')
    except LookupError:
        pass

    try:
        useroption_parsing = sip.get('general', 'legacy_useroption_parsing')[0]
        set_value('ignore_uri_user_options', useroption_parsing, 'global', pjsip, nmapped, 'global')
    except LookupError:
        pass

    try:
        timer_t1 = sip.get('general', 'timert1')[0]
        set_value('timer_t1', timer_t1, section, pjsip, nmapped, type)
    except LookupError:
        pass

    try:
        timer_b = sip.get('general', 'timerb')[0]
        set_value('timer_b', timer_b, section, pjsip, nmapped, type)
    except LookupError:
        pass

    try:
        compact_headers = sip.get('general', 'compactheaders')[0]
        set_value('compact_headers', compact_headers, section, pjsip, nmapped, type)
    except LookupError:
        pass


def convert(sip, filename, non_mappings, include):
    """
    Entry point for configuration file conversion. This
    function will create a pjsip.conf object and begin to
    map specific sections from sip.conf into it.
    Returns the new pjsip.conf object once completed
    """
    pjsip = sip.__class__()
    non_mappings[filename] = astdicts.MultiOrderedDict()
    nmapped = non_mapped(non_mappings[filename])
    if not include:
        # Don't duplicate transport and registration configs
        map_system(sip, pjsip, nmapped)
        map_transports(sip, pjsip, nmapped)
        map_registrations(sip, pjsip, nmapped)
    map_auth(sip, pjsip, nmapped)
    for section in sip.sections():
        if section == 'authentication':
            pass
        else:
            map_peer(sip, section, pjsip, nmapped)
            map_setvars(sip, section, pjsip, nmapped)

    find_non_mapped(sip.defaults(), nmapped)
    find_non_mapped(sip.sections(), nmapped)

    for key, val in sip.includes().iteritems():
        pjsip.add_include(PREFIX + key, convert(val, PREFIX + key,
                          non_mappings, True)[0])
    return pjsip, non_mappings


def write_pjsip(filename, pjsip, non_mappings):
    """
    Write pjsip.conf file to disk
    """
    try:
        with open(filename, 'wt') as fp:
            fp.write(';--\n')
            fp.write(';;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n')
            fp.write('Non mapped elements start\n')
            fp.write(';;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n\n')
            astconfigparser.write_dicts(fp, non_mappings[filename])
            fp.write(';;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n')
            fp.write('Non mapped elements end\n')
            fp.write(';;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n')
            fp.write('--;\n\n')
            # write out include file(s)
            pjsip.write(fp)

    except IOError:
        print("Could not open file " + filename + " for writing", file=sys.stderr)

###############################################################################


def cli_options():
    """
    Parse command line options and apply them. If invalid input is given,
    print usage information
    """
    global PREFIX
    global QUIET
    usage = "usage: %prog [options] [input-file [output-file]]\n\n" \
        "Converts the chan_sip configuration input-file to the chan_pjsip output-file.\n" \
        "The input-file defaults to 'sip.conf'.\n" \
        "The output-file defaults to 'pjsip.conf'."
    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-p', '--prefix', dest='prefix', default=PREFIX,
                      help='output prefix for include files')
    parser.add_option('-q', '--quiet', dest='quiet', default=False, action='store_true',
                      help="don't print messages to stdout")

    options, args = parser.parse_args()
    PREFIX = options.prefix
    if options.quiet:
        QUIET = True

    sip_filename = args[0] if len(args) else 'sip.conf'
    pjsip_filename = args[1] if len(args) == 2 else 'pjsip.conf'

    return sip_filename, pjsip_filename


def info(msg):
    if QUIET:
        return
    print(msg)


if __name__ == "__main__":
    sip_filename, pjsip_filename = cli_options()
    # configuration parser for sip.conf
    sip = astconfigparser.MultiOrderedConfigParser()
    info('Please, report any issue at:')
    info('    https://issues.asterisk.org/')
    info('Reading ' + sip_filename)
    sip.read(sip_filename)
    info('Converting to PJSIP...')
    pjsip, non_mappings = convert(sip, pjsip_filename, dict(), False)
    info('Writing ' + pjsip_filename)
    write_pjsip(pjsip_filename, pjsip, non_mappings)



#!/usr/bin/env python
"""Process a ref debug log

 This file will process a log file created by enabling
 the refdebug config option in asterisk.conf.

 See http://www.asterisk.org for more information about
 the Asterisk project. Please do not directly contact
 any of the maintainers of this project for assistance;
 the project provides a web site, mailing lists and IRC
 channels for your use.

 This program is free software, distributed under the terms of
 the GNU General Public License Version 2. See the LICENSE file
 at the top of the source tree.

 Copyright (C) 2014, Digium, Inc.
 Matt Jordan <mjordan@digium.com>
"""

from __future__ import print_function
import sys
import os

from optparse import OptionParser


def parse_line(line):
    """Parse out a line into its constituent parts.

    Keyword Arguments:
    line The line from a ref debug log to parse out

    Returns:
    A dictionary containing the options, or None
    """
    tokens = line.strip().split(',', 7)
    if len(tokens) < 8:
        print("ERROR: ref debug line '%s' contains fewer tokens than "
              "expected: %d" % (line.strip(), len(tokens)))
        return None

    processed_line = {'addr': tokens[0],
                      'delta': tokens[1],
                      'thread_id': tokens[2],
                      'file': tokens[3],
                      'line': tokens[4],
                      'function': tokens[5],
                      'state': tokens[6],
                      'tag': tokens[7],
                      }
    return processed_line


def process_file(options):
    """The routine that kicks off processing a ref file

    Keyword Arguments:
    filename The full path to the file to process

    Returns:
    A tuple containing:
        - A list of objects whose lifetimes were completed
            (i.e., finished objects)
        - A list of objects referenced after destruction
            (i.e., invalid objects)
        - A list of objects whose lifetimes were not completed
            (i.e., leaked objects)
        - A list of objects whose lifetimes are skewed
            (i.e., Object history starting with an unusual ref count)
    """

    finished_objects = []
    invalid_objects = []
    leaked_objects = []
    skewed_objects = []
    current_objects = {}
    filename = options.filepath

    with open(filename, 'r') as ref_file:
        for line in ref_file:
            parsed_line = parse_line(line)
            if not parsed_line:
                continue

            invalid = False
            obj = parsed_line['addr']

            if obj not in current_objects:
                current_objects[obj] = {'log': [], 'curcount': 1}
                if 'constructor' in parsed_line['state']:
                    # This is the normal expected case
                    pass
                elif 'invalid' in parsed_line['state']:
                    invalid = True
                    current_objects[obj]['curcount'] = 0
                    if options.invalid:
                        invalid_objects.append((obj, current_objects[obj]))
                elif 'destructor' in parsed_line['state']:
                    current_objects[obj]['curcount'] = 0
                    if options.skewed:
                        skewed_objects.append((obj, current_objects[obj]))
                else:
                    current_objects[obj]['curcount'] = int(
                        parsed_line['state'])
                    if options.skewed:
                        skewed_objects.append((obj, current_objects[obj]))
            else:
                current_objects[obj]['curcount'] += int(parsed_line['delta'])

            # Suppress object sizes and lock-state from output logs.
            if 'constructor' in parsed_line['state']:
                parsed_line['state'] = '**constructor**'
            elif 'destructor' in parsed_line['state']:
                parsed_line['state'] = '**destructor**'

            current_objects[obj]['log'].append(
                "[%s] %s:%s %s: %s %s - [%s]" % (
                    parsed_line['thread_id'],
                    parsed_line['file'],
                    parsed_line['line'],
                    parsed_line['function'],
                    parsed_line['delta'],
                    parsed_line['tag'],
                    parsed_line['state']))

            # It is possible for curcount to go below zero if someone
            # unrefs an object by two or more when there aren't that
            # many refs remaining.  This condition abnormally finishes
            # the object.
            if current_objects[obj]['curcount'] <= 0:
                if current_objects[obj]['curcount'] < 0:
                    current_objects[obj]['log'].append(
                        "[%s] %s:%s %s: %s %s - [%s]" % (
                            parsed_line['thread_id'],
                            parsed_line['file'],
                            parsed_line['line'],
                            parsed_line['function'],
                            "+0",
                            "Object abnormally finalized",
                            "**implied destructor**"))
                    # Highlight the abnormally finished object in the
                    # invalid section as well as reporting it in the normal
                    # finished section.
                    if options.invalid:
                        invalid_objects.append((obj, current_objects[obj]))
                if not invalid and options.normal:
                    finished_objects.append((obj, current_objects[obj]))
                del current_objects[obj]

    if options.leaks:
        for (key, lines) in current_objects.items():
            leaked_objects.append((key, lines))
    return (finished_objects, invalid_objects, leaked_objects, skewed_objects)


def print_objects(objects, prefix=""):
    """Prints out the objects that were processed

    Keyword Arguments:
    objects A list of objects to print
    prefix  A prefix to print that specifies something about
            this object
    """

    print("======== %s Objects ========" % prefix)
    print("\n")
    for obj in objects:
        print("==== %s Object %s history ====" % (prefix, obj[0]))
        for line in obj[1]['log']:
            print(line)
        print("\n")


def main(argv=None):
    """Main entry point for the script"""

    ret_code = 0

    if argv is None:
        argv = sys.argv

    parser = OptionParser()

    parser.add_option("-f", "--file", action="store", type="string",
                      dest="filepath", default="/var/log/asterisk/refs",
                      help="The full path to the refs file to process")
    parser.add_option("-i", "--suppress-invalid", action="store_false",
                      dest="invalid", default=True,
                      help="If specified, don't output invalid object "
                           "references")
    parser.add_option("-l", "--suppress-leaks", action="store_false",
                      dest="leaks", default=True,
                      help="If specified, don't output leaked objects")
    parser.add_option("-n", "--suppress-normal", action="store_false",
                      dest="normal", default=True,
                      help="If specified, don't output objects with a "
                           "complete lifetime")
    parser.add_option("-s", "--suppress-skewed", action="store_false",
                      dest="skewed", default=True,
                      help="If specified, don't output objects with a "
                           "skewed lifetime")

    (options, args) = parser.parse_args(argv)

    if not options.invalid and not options.leaks and not options.normal \
            and not options.skewed:
        print("All options disabled", file=sys.stderr)
        return -1

    if not os.path.isfile(options.filepath):
        print("File not found: %s" % options.filepath, file=sys.stderr)
        return -1

    try:
        (finished_objects,
         invalid_objects,
         leaked_objects,
         skewed_objects) = process_file(options)

        if options.invalid and len(invalid_objects):
            print_objects(invalid_objects, "Invalid Referenced")
            ret_code |= 4

        if options.leaks and len(leaked_objects):
            print_objects(leaked_objects, "Leaked")
            ret_code |= 1

        if options.skewed and len(skewed_objects):
            print_objects(skewed_objects, "Skewed")
            ret_code |= 2

        if options.normal:
            print_objects(finished_objects, "Finalized")

    except (KeyboardInterrupt, SystemExit, IOError):
        print("File processing cancelled", file=sys.stderr)
        return -1

    return ret_code


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Copyright (C) 2016, Digium, Inc.

This program is free software, distributed under the terms of
the GNU General Public License Version 2.
"""

import re
import glob
import itertools

from astdicts import OrderedDict
from astdicts import MultiOrderedDict


def merge_values(left, right, key):
    """Merges values from right into left."""
    if isinstance(left, list):
        vals0 = left
    else:  # assume dictionary
        vals0 = left[key] if key in left else []
    vals1 = right[key] if key in right else []

    return vals0 + [i for i in vals1 if i not in vals0]

###############################################################################


class Section(MultiOrderedDict):
    """
    A Section is a MultiOrderedDict itself that maintains a list of
    key/value options.  However, in the case of an Asterisk config
    file a section may have other defaults sections that is can pull
    data from (i.e. templates).  So when an option is looked up by key
    it first checks the base section and if not found looks in the
    added default sections. If not found at that point then a 'KeyError'
    exception is raised.
    """
    count = 0

    def __init__(self, defaults=None, templates=None):
        MultiOrderedDict.__init__(self)
        # track an ordered id of sections
        Section.count += 1
        self.id = Section.count
        self._defaults = [] if defaults is None else defaults
        self._templates = [] if templates is None else templates

    def __cmp__(self, other):
        """
        Use self.id as means of determining equality
        """
        return (self.id > other.id) - (self.id < other.id)

    def __eq__(self, other):
        """
        Use self.id as means of determining equality
        """
        return self.id == other.id

    def __lt__(self, other):
        """
        Use self.id as means of determining equality
        """
        return self.id < other.id

    def __gt__(self, other):
        """
        Use self.id as means of determining equality
        """
        return self.id > other.id

    def __le__(self, other):
        """
        Use self.id as means of determining equality
        """
        return self.id <= other.id

    def __ge__(self, other):
        """
        Use self.id as means of determining equality
        """
        return self.id >= other.id

    def get(self, key, from_self=True, from_templates=True,
            from_defaults=True):
        """
        Get the values corresponding to a given key. The parameters to this
        function form a hierarchy that determines priority of the search.
        from_self takes priority over from_templates, and from_templates takes
        priority over from_defaults.

        Parameters:
        from_self - If True, search within the given section.
        from_templates - If True, search in this section's templates.
        from_defaults - If True, search within this section's defaults.
        """
        if from_self and key in self:
            return MultiOrderedDict.__getitem__(self, key)

        if from_templates:
            if self in self._templates:
                return []
            for t in self._templates:
                try:
                    # fail if not found on the search - doing it this way
                    # allows template's templates to be searched.
                    return t.get(key, True, from_templates, from_defaults)
                except KeyError:
                    pass

        if from_defaults:
            for d in self._defaults:
                try:
                    return d.get(key, True, from_templates, from_defaults)
                except KeyError:
                    pass

        raise KeyError(key)

    def __getitem__(self, key):
        """
        Get the value for the given key. If it is not found in the 'self'
        then check inside templates and defaults before declaring raising
        a KeyError exception.
        """
        return self.get(key)

    def keys(self, self_only=False):
        """
        Get the keys from this section. If self_only is True, then
        keys from this section's defaults and templates are not
        included in the returned value
        """
        res = MultiOrderedDict.keys(self)
        if self_only:
            return res

        for d in self._templates:
            for key in d.keys():
                if key not in res:
                    res.append(key)

        for d in self._defaults:
            for key in d.keys():
                if key not in res:
                    res.append(key)
        return res

    def add_defaults(self, defaults):
        """
        Add a list of defaults to the section. Defaults are
        sections such as 'general'
        """
        defaults.sort()
        for i in defaults:
            self._defaults.insert(0, i)

    def add_templates(self, templates):
        """
        Add a list of templates to the section.
        """
        templates.sort()
        for i in templates:
            self._templates.insert(0, i)

    def get_merged(self, key):
        """Return a list of values for a given key merged from default(s)"""
        # first merge key/values from defaults together
        merged = []
        for i in reversed(self._defaults):
            if not merged:
                merged = i
                continue
            merged = merge_values(merged, i, key)

        for i in reversed(self._templates):
            if not merged:
                merged = i
                continue
            merged = merge_values(merged, i, key)

        # then merge self in
        return merge_values(merged, self, key)

###############################################################################

COMMENT = ';'
COMMENT_START = ';--'
COMMENT_END = '--;'

DEFAULTSECT = 'general'


def remove_comment(line, is_comment):
    """Remove any commented elements from the line."""
    if not line:
        return line, is_comment

    if is_comment:
        part = line.partition(COMMENT_END)
        if part[1]:
            # found multi-line comment end check string after it
            return remove_comment(part[2], False)
        return "", True

    part = line.partition(COMMENT_START)
    if part[1] and not part[2].startswith('-'):
        # found multi-line comment start check string before
        # it to make sure there wasn't an eol comment in it
        has_comment = part[0].partition(COMMENT)
        if has_comment[1]:
            # eol comment found return anything before it
            return has_comment[0], False

        # check string after it to see if the comment ends
        line, is_comment = remove_comment(part[2], True)
        if is_comment:
            # return possible string data before comment
            return part[0].strip(), True

        # otherwise it was an embedded comment so combine
        return ''.join([part[0].strip(), ' ', line]).rstrip(), False

    # find the first occurence of a comment that is not escaped
    match = re.match(r'.*?([^\\];)', line)

    if match:
         # the end of where the real string is is where the comment starts
         line = line[0:(match.end()-1)]
    if line.startswith(";"):
         # if the line is actually a comment just ignore it all
         line = ""

    return line.replace("\\", "").strip(), False

def try_include(line):
    """
    Checks to see if the given line is an include.  If so return the
    included filename, otherwise None.
    """

    match = re.match('^#include\s*([^;]+).*$', line)
    if match:
        trimmed = match.group(1).rstrip()
        quoted = re.match('^"([^"]+)"$', trimmed)
        if quoted:
            return quoted.group(1)
        bracketed = re.match('^<([^>]+)>$', trimmed)
        if bracketed:
            return bracketed.group(1)
        return trimmed
    return None


def try_section(line):
    """
    Checks to see if the given line is a section. If so return the section
    name, otherwise return 'None'.
    """
    # leading spaces were stripped when checking for comments
    if not line.startswith('['):
        return None, False, []

    section, delim, templates = line.partition(']')
    if not templates:
        return section[1:], False, []

    # strip out the parens and parse into an array
    templates = templates.replace('(', "").replace(')', "").split(',')
    # go ahead and remove extra whitespace
    templates = [i.strip() for i in templates]
    try:
        templates.remove('!')
        return section[1:], True, templates
    except:
        return section[1:], False, templates


def try_option(line):
    """Parses the line as an option, returning the key/value pair."""
    data = re.split('=>?', line, 1)
    # should split in two (key/val), but either way use first two elements
    return data[0].rstrip(), data[1].lstrip()

###############################################################################


def find_dict(mdicts, key, val):
    """
    Given a list of mult-dicts, return the multi-dict that contains
    the given key/value pair.
    """

    def found(d):
        return key in d and val in d[key]

    try:
        return [d for d in mdicts if found(d)][0]
    except IndexError:
        raise LookupError("Dictionary not located for key = %s, value = %s"
                          % (key, val))


def write_dicts(config_file, mdicts):
    """Write the contents of the mdicts to the specified config file"""
    for section, sect_list in mdicts.iteritems():
        # every section contains a list of dictionaries
        for sect in sect_list:
            config_file.write("[%s]\n" % section)
            for key, val_list in sect.iteritems():
                # every value is also a list
                for v in val_list:
                    key_val = key
                    if v is not None:
                        key_val += " = " + str(v)
                        config_file.write("%s\n" % (key_val))
            config_file.write("\n")

###############################################################################


class MultiOrderedConfigParser:
    def __init__(self, parent=None):
        self._parent = parent
        self._defaults = MultiOrderedDict()
        self._sections = MultiOrderedDict()
        self._includes = OrderedDict()

    def find_value(self, sections, key):
        """Given a list of sections, try to find value(s) for the given key."""
        # always start looking in the last one added
        sections.sort(reverse=True)
        for s in sections:
            try:
                # try to find in section and section's templates
                return s.get(key, from_defaults=False)
            except KeyError:
                pass

        # wasn't found in sections or a section's templates so check in
        # defaults
        for s in sections:
            try:
                # try to find in section's defaultsects
                return s.get(key, from_self=False, from_templates=False)
            except KeyError:
                pass

        raise KeyError(key)

    def defaults(self):
        return self._defaults

    def default(self, key):
        """Retrieves a list of dictionaries for a default section."""
        return self.get_defaults(key)

    def add_default(self, key, template_keys=None):
        """
        Adds a default section to defaults, returning the
        default Section object.
        """
        if template_keys is None:
            template_keys = []
        return self.add_section(key, template_keys, self._defaults)

    def sections(self):
        return self._sections

    def section(self, key):
        """Retrieves a list of dictionaries for a section."""
        return self.get_sections(key)

    def get_sections(self, key, attr='_sections', searched=None):
        """
        Retrieve a list of sections that have values for the given key.
        The attr parameter can be used to control what part of the parser
        to retrieve values from.
        """
        if searched is None:
            searched = []
        if self in searched:
            return []

        sections = getattr(self, attr)
        res = sections[key] if key in sections else []
        searched.append(self)
        if self._includes:
            res.extend(list(itertools.chain(*[
                incl.get_sections(key, attr, searched)
                for incl in self._includes.itervalues()])))
        if self._parent:
            res += self._parent.get_sections(key, attr, searched)
        return res

    def get_defaults(self, key):
        """
        Retrieve a list of defaults that have values for the given key.
        """
        return self.get_sections(key, '_defaults')

    def add_section(self, key, template_keys=None, mdicts=None):
        """
        Create a new section in the configuration. The name of the
        new section is the 'key' parameter.
        """
        if template_keys is None:
            template_keys = []
        if mdicts is None:
            mdicts = self._sections
        res = Section()
        for t in template_keys:
            res.add_templates(self.get_defaults(t))
        res.add_defaults(self.get_defaults(DEFAULTSECT))
        mdicts.insert(0, key, res)
        return res

    def includes(self):
        return self._includes

    def add_include(self, filename, parser=None):
        """
        Add a new #include file to the configuration.
        """
        if filename in self._includes:
            return self._includes[filename]

        self._includes[filename] = res = \
            MultiOrderedConfigParser(self) if parser is None else parser
        return res

    def get(self, section, key):
        """Retrieves the list of values from a section for a key."""
        try:
            # search for the value in the list of sections
            return self.find_value(self.section(section), key)
        except KeyError:
            pass

        try:
            # section may be a default section so, search
            # for the value in the list of defaults
            return self.find_value(self.default(section), key)
        except KeyError:
            raise LookupError("key %r not found for section %r"
                              % (key, section))

    def multi_get(self, section, key_list):
        """
        Retrieves the list of values from a section for a list of keys.
        This method is intended to be used for equivalent keys. Thus, as soon
        as any match is found for any key in the key_list, the match is
        returned. This does not concatenate the lookups of all of the keys
        together.
        """
        for i in key_list:
            try:
                return self.get(section, i)
            except LookupError:
                pass

        # Making it here means all lookups failed.
        raise LookupError("keys %r not found for section %r" %
                          (key_list, section))

    def set(self, section, key, val):
        """Sets an option in the given section."""
        # TODO - set in multiple sections? (for now set in first)
        # TODO - set in both sections and defaults?
        if section in self._sections:
            self.section(section)[0][key] = val
        else:
            self.defaults(section)[0][key] = val

    def read(self, filename, sect=None):
        """Parse configuration information from a file"""
        try:
            with open(filename, 'rt') as config_file:
                self._read(config_file, sect)
        except IOError:
            print("Could not open file " + filename + " for reading")

    def _read(self, config_file, sect):
        """Parse configuration information from the config_file"""
        is_comment = False  # used for multi-lined comments
        for line in config_file:
            line, is_comment = remove_comment(line, is_comment)
            if not line:
                # line was empty or was a comment
                continue

            include_name = try_include(line)
            if include_name:
                for incl in sorted(glob.iglob(include_name)):
                    parser = self.add_include(incl)
                    parser.read(incl, sect)
                continue

            section, is_template, templates = try_section(line)
            if section:
                if section == DEFAULTSECT or is_template:
                    sect = self.add_default(section, templates)
                else:
                    sect = self.add_section(section, templates)
                continue

            key, val = try_option(line)
            if sect is None:
                raise Exception("Section not defined before assignment")
            sect[key] = val

    def write(self, config_file):
        """Write configuration information out to a file"""
        try:
            for key, val in self._includes.iteritems():
                val.write(key)
                config_file.write('#include "%s"\n' % key)

            config_file.write('\n')
            write_dicts(config_file, self._defaults)
            write_dicts(config_file, self._sections)
        except:
            try:
                with open(config_file, 'wt') as fp:
                    self.write(fp)
            except IOError:
                print("Could not open file " + config_file + " for writing")



from astconfigparser import MultiOrderedConfigParser

try:
    import pymysql as MySQLdb
    MySQLdb.install_as_MySQLdb()
except ImportError:
    # MySQLdb is compatible with Python 2 only.  Try it as a
    # fallback if pymysql is unavailable.
    import MySQLdb

import traceback

class SqlConfigParser(MultiOrderedConfigParser):

    _tablename = "sippeers"

    def __init__(self,tablename="sippeers"):
        self._tablename=tablename
        MultiOrderedConfigParser.__init__(self)

    def connect(self, user, password, host, port, database):
        self.cnx = MySQLdb.connect(user=user,passwd=password,host=host,port=int(port),db=database)

    def read(self, filename, sect=None):
        MultiOrderedConfigParser.read(self, filename, sect)
        # cursor = self.cnx.cursor(dictionary=True)
        cursor = self.cnx.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        cursor.execute("SELECT * from `" + MySQLdb.escape_string(self._tablename) + "`")
        rows = cursor.fetchall()

        for row in rows:
            sect = self.add_section(row['name'])
            for key in row:
                if (row[key] != None):
                    for elem in str(row[key]).split(";"):
                        sect[key] = elem
                    #sect[key] = str(row[key]).split(";")

    def write_dicts(self, config_file, mdicts):
        """Write the contents of the mdicts to the specified config file"""
        for section, sect_list in mdicts.iteritems():
            # every section contains a list of dictionaries
            for sect in sect_list:
                sql = "INSERT INTO "
                if (sect.get('type')[0] == "endpoint"):
                    sql += "ps_endpoints "
                elif (sect.get('type')[0] == "aor" and section != "sbc"):
                    sql += "ps_aors "
                elif (sect.get('type')[0] == "identify"):
                    sql += "ps_endpoint_id_ips"
                else:
                    continue

                sql += " SET `id` = " + "\"" + MySQLdb.escape_string(section) + "\""
                for key, val_list in sect.iteritems():
                    if key == "type":
                        continue
                    # every value is also a list

                    key_val = " `" + key + "`"
                    key_val += " = " + "\"" + MySQLdb.escape_string(";".join(val_list)) + "\""
                    sql += ","
                    sql += key_val

                config_file.write("%s;\n" % (sql))

    def write(self, config_file):
        """Write configuration information out to a file"""
        try:
            self.write_dicts(config_file, self._sections)
        except:
                print("Could not open file " + config_file + " for writing")
                traceback.print_exc()



#!/usr/local/bin/python3.13

from sip_to_pjsip import convert
import sip_to_pjsip
import optparse


import sqlconfigparser


def write_pjsip(filename, pjsip, non_mappings):
    """
    Write pjsip.sql file to disk
    """
    try:
        with open(filename, 'wt') as fp:
            pjsip.write(fp)

    except IOError:
        print("Could not open file " + filename + " for writing")

def cli_options():
    """
    Parse command line options and apply them. If invalid input is given,
    print usage information

    """
    global user
    global password
    global host
    global port
    global database
    global table

    usage = "usage: %prog [options] [input-file [output-file]]\n\n" \
        "Converts the chan_sip configuration input-file to mysql output-file.\n" \
        "The input-file defaults to 'sip.conf'.\n" \
        "The output-file defaults to 'pjsip.sql'."
    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-u', '--user', dest='user', default="root",
                      help='mysql username')
    parser.add_option('-p', '--password', dest='password', default="root",
                      help='mysql password')
    parser.add_option('-H', '--host', dest='host', default="127.0.0.1",
                      help='mysql host ip')
    parser.add_option('-P', '--port', dest='port', default="3306",
                      help='mysql port number')
    parser.add_option('-D', '--database', dest='database', default="asterisk",
                      help='mysql port number')
    parser.add_option('-t', '--table', dest='table', default="sippeers",
                      help='name of sip realtime peers table')

    options, args = parser.parse_args()

    user = options.user
    password = options.password
    host = options.host
    port = options.port
    database = options.database
    table = options.table

    sip_filename = args[0] if len(args) else 'sip.conf'
    pjsip_filename = args[1] if len(args) == 2 else 'pjsip.sql'

    return sip_filename, pjsip_filename

if __name__ == "__main__":
    sip_filename, pjsip_filename = cli_options()
    sip = sqlconfigparser.SqlConfigParser(table)
    sip_to_pjsip.sip = sip
    sip.connect(user,password,host,port,database)
    print('Please, report any issue at:')
    print('    https://issues.asterisk.org/')
    print('Reading ' + sip_filename)
    sip.read(sip_filename)
    print('Converting to PJSIP realtime sql...')
    pjsip, non_mappings = convert(sip, pjsip_filename, dict(), False)
    print('Writing ' + pjsip_filename)
    write_pjsip(pjsip_filename, pjsip, non_mappings)

